@@ -11,7 +11,7 @@ use warnings;
 use base qw( IO::Async::Notifier );
 IO::Async::Notifier->VERSION( '0.63' ); # adopt_future
 
-our $VERSION = '0.15';
+our $VERSION = '0.16';
 $VERSION = eval $VERSION;
 
 use Carp;
@@ -21,6 +21,7 @@ use Future::Utils qw( repeat );
 use JSON qw( encode_json decode_json );
 
 use Data::Dump 'pp';
+use File::stat;
 use Struct::Dumb;
 use Time::HiRes qw( time );
 use URI;
@@ -306,6 +307,62 @@ sub _do_DELETE
    );
 }
 
+sub _do_POST_file
+{
+   my $self = shift;
+   my ( $path, %params ) = @_;
+
+   my $uri = $self->_uri_for_path( "" );
+   $uri->path( "/_matrix" . $path );
+
+   my $req = HTTP::Request->new( "POST" , $uri );
+   $req->header( Content_type => $params{content_type} );
+
+   my $body;
+
+   if( defined $params{content} ) {
+      $req->content( $params{content} );
+      $req->header( Content_length => length $req->content );
+   }
+   elsif( defined $params{file} or defined $params{fh} ) {
+      my $fh = $params{fh};
+      $fh or open $fh, "<", $params{file} or
+         return Future->fail( "Cannot read $params{file} - $!", open => );
+
+      $body = sub {
+         $fh->read( my $buffer, 65536 ) or return undef;
+         return $buffer;
+      };
+
+      $req->header( Content_length => $params{content_length} // ( stat $fh )->size );
+   }
+
+   my $f = $self->{ua}->do_request(
+      request => $req,
+      request_body => $body,
+   )->then( sub {
+      my ( $response ) = @_;
+      $response->content_type eq "application/json" or
+         return Future->fail( "Expected application/json response", matrix => );
+
+      my $content = $response->content;
+      my $uri;
+      if( length $content and $content ne q("") ) {
+         eval {
+            $content = decode_json( $content );
+            1;
+         } or  
+            return Future->fail( "Unable to parse JSON response " );
+         return Future->done( $content, $response );
+      }
+      else {
+         return Future->done( undef, $response );
+      }
+   });
+
+   return $self->adopt_future( $f );
+}
+
 =head2 $matrix->login( %params )->get
 
 Performs the necessary steps required to authenticate with the configured
@@ -914,6 +971,67 @@ sub delete_alias
       ->then_done();
 }
 
+=head2 $content_uri = $matrix->upload( %params )->get
+
+Performs a post to the server's media content repository, to upload a new
+piece of content, returning the content URI that points to it.
+
+The content can be specified in any of three ways, with the following three
+mutually-exclusive arguments:
+
+=over 4
+
+=item content => STRING
+
+Gives the content directly as an immediate scalar value.
+
+=item file => STRING
+
+Gives the path to a readable file on the filesystem containing the content.
+
+=item fh => IO
+
+Gives an opened IO handle the content can be read from.
+
+=back
+
+The following additional arguments are also recognised:
+
+=over 4
+
+=item content_type => STRING
+
+Gives the MIME type of the content data.
+
+=item content_length => INT
+
+Optional. If the content is being delivered from an opened filehandle (via the
+C<fh> argument), this gives the total length in bytes. This is required in
+cases such as reading from pipes, when the length of the content isn't
+immediately available such as by C<stat()>ing the filehandle.
+
+=back
+
+=cut
+
+sub upload
+{
+   my $self = shift;
+   my %params = @_;
+
+   defined $params{content_type} or
+      croak "Require 'content_type'";
+
+   defined $params{content} or defined $params{file} or defined $params{fh} or
+      croak "Require 'content', 'file' or 'fh'";
+
+   # This one takes ~full URL paths
+   $self->_do_POST_file( "/media/v1/upload", %params )->then( sub {
+      my ( $content, $response ) = @_;
+      Future->done( $content->{content_uri} );
+   });
+}
+
 ## Incoming events
 
 sub _handle_event_m_presence
